"""Class definitions to apply the cemorization."""

from datetime import datetime
from getpass import getuser
import json
from pathlib import Path
import os
import re
import shlex
from subprocess import run, PIPE
import sys
from tempfile import TemporaryDirectory, NamedTemporaryFile

from jinja2 import Template
from netCDF4 import num2date, date2num, Dataset
import numpy as np

SCRATCH_DIR = Path('/scratch/') / getuser()[0] / getuser()

from .utils import icon2datetime, units

class CMoriser:
    """Collection of methods to cmorise arbitrary input data."""

    def __init__(self, drstable, runinfo, cmortable):
        """Create an instance of the Cmorise class using meta dta."""

        self.drstable = drstable
        self.runinfo = runinfo
        self.cmortable = cmortable
        tmpl_dir = Path(__file__).parent.absolute() / 'cmor'
        cv_file = tmpl_dir / self.runinfo['_control_vocabulary_file']
        with cv_file.open() as f:
            self.cv_table = json.load(f)['CV']

    @staticmethod
    def _is_length_dim(inunit):
        """Check if a given unit is a length only dimension."""

        return [dim[1:-1] for dim in inunit.units.dimensionality] == ['length']


    def get_unit_conversion_factor(self, inputunit, outputunit, dt):
        """Get a conversion factor between to units

           Parameters:
           ===========

            inputunit : str
                Input unit that needs to be converted
            outputunit : str
                Target unit to convert input unit to
            dt : datetime.timedelta
                Timedelata object for two consecutive timesteps

            Returns:
            ========
             None : The netCDF4.Dataset object is modified inplace
        """

        unit = [inputunit, outputunit]
        # Try to convert the units to pint.units objects
        for n, uu in enumerate(unit):
            if 'frac' in unit[n] or unit[n] in ('1', '1.'):
                # Non-units assignment
                uu = ''
            elif '%' in unit[n]:
                uu = 'percent'
            else:
                # Pint cannot deal with units like s-1
                # try to identify those units and replace numeric values
                # with a '^' (if numeric characters are present but no
                # power sign is given - like in m s-1 -> m s^-1)
                for i in re.findall("[-\d]+", unit[n]):
                    if not '^' in unit[n] or '**' in unit[n]:
                        uu = uu.replace(i, f'^{i}')
            unit[n] = units(uu)
        inunit, outunit = unit
        if inunit == outunit:
            # If the units are euqal we do not need to do anything
            return 1
        rho_w = 1000 * units('kg/m^3') # ~ Density of water
        density_unit = units('kg/m**2').units
        if self._is_length_dim(inunit) and outunit.units == density_unit:
            # Most likely this is a water based variable, hence convert
            # a water hight unit to a density unit using the density of water
            return (inunit.to_base_units() * rho_w).magnitude
        if self._is_length_dim(outunit) and inunit.units == density_unit:
            # Same as above but vice versa
            conv = outunit.to_base_units() / rho_w
            return conv.to(inunit).magnitude #Convert from SI to base units
        mass_flux_unit = units('kg/m**2/s').units
        flux = (rho_w / (dt.total_seconds() * units('s'))).to_base_units()
        if self._is_length_dim(inunit) and outunit.units == mass_flux_unit:
            # Most likely this is a water based variable that needs to be
            # converted to flux unit by aplying rho_w and the time difference
            return (flux*inunit).to_base_units().magnitude
        if self._is_length_dim(outunit) and inunit.units == mass_flux_unit:
            # Same flux cas as above, just vice versa
            return ((outunit.to_base_units() / flux).to(inunit)).magnitude
        # Do a 'normal' magnitude conversion
        return inunit.to(outunit).magnitude

    def correct_file(self, infile, outfile, varname, lookup=None):
        """Run a first correction of the file.

        Parameters:
        ===========

        infile : str, pathlib.Path
            Input file that needs to be corrected (will be don inplace)
        varname : str
            Name of the input variable
        lookup : dict, pd.DataFrame
            Variable information of the stored file
        """
        with Dataset(str(infile), 'a') as f:
            tunits = 'seconds since 1970-01-01 00:00:00'
            try:
                datetimes = num2date(f['time'][:], f['time'].units)
            except ValueError:
                datetimes = icon2datetime(f['time'][:])
            var_metadata = self.cmortable['variable_entry'][varname]
            unit = lookup.get('units') or f[varnname].units
            inunit = unit.strip('[').strip(']')
            outunit = var_metadata['units']
            long_name = lookup.get('long_name') or f[varname].long_name
            f.variables['time'][:] = date2num(datetimes, tunits)
            f.variables['time'].units = tunits
            f.variables['time'].axis = 'T'
            f.variables['time'].long_name = f['time'].out_name = 'time'
            f.variables[varname].long_name = var_metadata['long_name']
            f.variables[varname].units = outunit
            f.variables[varname].standard_name = var_metadata['standard_name']
            dt = datetimes[1] - datetimes[0]
            mul = self.get_unit_conversion_factor(inunit, outunit, dt)
            f.variables[varname][:] = mul * f.variables[varname][:]
            for attr, value in self.get_global_attrs().items():
                setattr(f, attr, value)
        self.correct_metadata(infile, outfile, varname,
                              long_name=long_name, units=unit)

    def _get_drs_attrs(self, varname, timestamp):
        """Construct all drs attributes."""

        lookup = { k: Template(v).render(**self.runinfo)\
                for (k, v) in self.cv_table['path_facets'].items() }
        lookup['variable'] = varname
        lookup['varname'] = varname,
        lookup['version'] = timestamp.strftime('v%Y%m%d')
        lookup['project'] = lookup['project'].lower()
        return lookup

    def _get_timestamp_from_file(self, inpfile):

        cmd = shlex.split(f'cdo showtimestamp {inpfile}')
        res = run(cmd, stdout=PIPE, stderr=PIPE, check=True)
        out = res.stdout.decode('utf-8').strip().split()
        timetamps = ' '.join(out).split(' ')
        ts = [datetime.fromisoformat(ts).strftime('%Y%m%d%M%d')\
                for ts in (timetamps[0], timetamps[-1])]
        return '-'.join(ts)


    def _get_filename_from_struct(self, varname, timestamp, inpfile):
        """Construct the output filename from the DRS structure."""

        key = tuple(self.drstable.keys())[0]
        file_dir = SCRATCH_DIR / key

        file_name, sub_dir = [], []
        drs_struct = self._get_drs_attrs(varname, timestamp)
        ts = self._get_timestamp_from_file(inpfile)
        for part in self.drstable[key]['parts_file_name']:
            if part == 'time':
                file_name.append(ts)
            else:
                file_name.append(drs_struct[part])
        for part in self.drstable[key]['parts_dir']:
            if part != 'file_name':
                sub_dir.append(drs_struct[part])
        fname = Path('/'.join(sub_dir)) / '_'.join(file_name)
        return file_dir / fname.with_suffix('.nc')


    def get_global_attrs(self):
        """Get all global attributes from the experiment info file."""

        keys = {}
        for key in self.cv_table['required_global_attributes']:
            try:
                keys[key] = self.runinfo[key]
            except KeyError:
                raise KeyError(f'Key : {key} expected to be in run control file.')
        self.cv_table['keys'].setdefault('initialization_method', 1)
        self.cv_table['keys'].setdefault('physics_version', 1)
        global_attrs = { k: Template(v).render(**self.runinfo) for (k, v) in self.cv_table['keys'].items() }
        global_attrs['institution'] = self.cv_table['institutions'][keys['institution_id']]
        return global_attrs


    def correct_metadata(self, filename, outfile, inputVarName,
                         long_name='', units=''):
        """Apply metadata correction using cdo cmorlite.

        Parameters:
        ===========

         filename : str, pathlib.Path
            Input filename
         outfile : str, pathlib.Path
            Name of the target output file
         inputVarName : str
            Name of the variable
         long_name : str, (default: None)
            Change the long_name attribute of the filename accordingly
            if None given (default) the long_name attr. from the metadata
            lookup file is taken.
         units : str, (default: None)
            Change the units attribute of the filename accordingly
            if None given (default) the units attr. from the metadata
            lookup file is taken.
        """

        lookup = self.cmortable['variable_entry'][inputVarName]
        long_name = lookup.get('long_name') or long_name
        units = lookup.get('units') or units
        with NamedTemporaryFile(suffix='.txt') as metafile:
            with open(metafile.name, 'w') as f:
                meta_data = dict(
                        name=inputVarName,
                        out_name=inputVarName,
                        long_name=long_name,
                        units=units,
                        factor=1.0,
                        missing_value=1e20
                        )
                meta_string='&parameter '+' '.join([f'{k}={v}' for (k, v) in meta_data.items()])
                f.write(meta_string+' /')
            cmd = f'cdo -O -s -f nc4 -z zip_6 --cmor -cmorlite,{metafile.name} {filename} {outfile}'
            res = run(shlex.split(cmd), stderr=PIPE, stdout=PIPE, check=True)

def _import(string, table_name):
    """Import the DRSFile structure definition from freva's file.py."""

    with TemporaryDirectory() as td:
        sys.path.insert(0, td)
        with (Path(td) / 'file.py').open('w') as f:
            f.write(string)
        from file import DRSFile
        sys.path.pop(0)
    try:
        struct =  DRSFile.DRS_STRUCTURE[table_name]
    except KeyError:
        raise KeyError(f'table name not found: {table_name}')
    except NameError:
        raise ValueError('freva module not loaded')
    del DRSFile
    return {table_name:struct}


def get_cmor_structure(table_name):
    """Get the DRS structure from freva's file.py

    This method tries to import freva's file module from the
    config4freva path, regardless of the python environment.
    To be able to make use of this method make sure the freva module is loaded.

    Parameters:
    ===========

     table_name: str
        The name of the DRS definition in the file module

    Returns:
    ========
    dict : dictionary containing the DRS file structure definitions.
    """

    try:
        freva_config = tuple(Path(os.environ['PYTHONPATH']).parents)[1]
    except KeyError:
        raise ValueError('freva module not loaded')
    freva_config = freva_config / 'misc4freva' / 'config4freva' / 'file.py'
    try:
        with freva_config.open() as f:
            cfg = f.readlines()[:]
    except FileNotFoundError:
        raise ValueError('freva module not loaded?')

    cfg  = '\n'.join([line.strip('\n') for line in cfg if 'evaluation_system' not in line])
    return _import(cfg, table_name)

