"""Entry point to setup and run the cemorization."""
import argparse
import itertools
import logging
import multiprocessing as mp
from pathlib import Path
import json
import sys

import pandas as pd
import toml

from .utils import read_cmor_table, get_filenames, units
from .preproc import preproc
from ._cmorize import get_cmor_structure


__all__ = ('get_cmor_structure', 'get_filenames', 'preproc', 'read_cmor_table',
           'prun', 'read_data_from_runinfo', 'units')


def parse_args(argv=None):
    """Consturct command line argument parser."""

    ap = argparse.ArgumentParser(prog='py-cmorlite',
            description="""Cmorize a set of files""",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument('metadata', metavar='metadata', type=Path,
                    help='Path to json or toml meta-data file format')
    ap.add_argument('input', metavar='input', type=Path, nargs="+",
                    help='Input directory where to find files')
    ap.add_argument('--file-type', type=str, default='.nc',
                   help='Filetypes to look for in the folder')
    ap.add_argument('--glob-pattern', type=str, default='*',
                   help='Glob pattern to be used when searching for files')
    ap.add_argument('--overwrite', action='store_true', default=False,
                   help='Overwrite existing files.')
    ap.add_argument('--model', type=str, default='remo',
                  help='Model output type that is going to be cmorised.')
    args = ap.parse_args()
    rundata, table = read_data_from_runinfo(args.metadata)
    return dict(
            rundata=rundata,
            cmortable=table,
            drstype=get_cmor_structure(rundata['_DRS_type']),
            files=get_filenames(args.input, args.file_type, args.glob_pattern),
            model=args.model,
            overwrite=args.overwrite
            )

def read_data_from_runinfo(info_json):
    """Get all information form the run inof json file.

    Parameters:
    ===========
     info_json : str, pathlib.Path
        path to the json file containing the experiment metadata

    Returns:
    ========
     dict, dict : dictionaries containing experiment information,
                  and variable target metadata information
    """

    # First the read the json file, with experiment information itself
    rundata = read_cmor_table(info_json, use_cmor_dir=False)
    return rundata, read_cmor_table(rundata['_variable_entry_file'])

def prun(processor, itterator):
    """Apply a function in parallel.

    Parameters:
    ===========
        processor, func
            function that is applied in parallel
        itterator, hashable
            itterator that is applied to the proccessing pool map

    Returns:
    ========
    list : results of the functions
    """

    num_procs = int(max(.25*mp.cpu_count(), 4))
    with mp.Pool(processes=num_procs) as pool:
        itt = pool.map(processor, itterator)
        results = []
        for res in itt:
            results.append(res)
    return results

def _run(*args):
    """Function for that serves as entry point."""
    config = parse_args(sys.argv)
    args = config.pop('drstype'), config.pop('rundata'), config.pop('cmortable')
    Processor = preproc(*args, **config)
    prun(Processor.process, config['files'])

if __name__ == '__main__':
    _run()
