"""Collection of methods to apply various pre-processing methods.

Each model input demands different pre-processing strategies, this should be
reflected in the preprocessing defintions for each model.
"""

from datetime import datetime
from pathlib import Path
import shlex
from subprocess import run, PIPE
from tempfile import TemporaryDirectory
import pandas as pd

from ._cmorize import CMoriser, SCRATCH_DIR

TABLE_DIR = Path(__file__).parent / 'var_tables'

def preproc(*args, model='remo', **kwargs):
    """Preprocessor wrapper.
    This method will retrun a pre-processing method accoring to a model type

    Parameters:
    ===========
        args : tuple
            drstable, runifo, cmortable for the current run
        model : str, (default: remo)
            the input model name that is processed
        kwargs: additional keyword arguments passed

    Returns:
    ========
        DataProcessor.from_method: Preprocessing method
    """
    return DataProcessor.from_method(model, *args, **kwargs)

class DataProcessor(CMoriser):

    @classmethod
    def from_method(cls, model, drstable, runinfo, cmortable, **kwargs):

        method = f'_proc_{model}'
        cls.process = getattr(cls, method)
        return cls(drstable, runinfo, cmortable, model, **kwargs)

    def __init__(self, drstable, runinfo, cmortable, model, overwrite=False):
        """Inherit from CMoriser."""

        super().__init__(drstable, runinfo, cmortable)
        self.overwrite = overwrite
        var_table = TABLE_DIR / f'{model}.csv'
        self.var_lookup = pd.read_csv(var_tables, index_col=0)

    def _proc_remo(self, infile):
        """Process remo model output."""

        infile = Path(infile)
        if not infile.is_file():
            return 0
        temp_file = SCRATCH_DIR / infile.name
        var_num = int(infile.name.split('_')[1][1:])
        try:
            info = self.var_lookup.loc[var_num]
        except KeyError:
            return 0
        variable = info['cmorname']
        mtime = datetime.fromtimestamp(infile.stat().st_mtime)
        out_file = self._get_filename_from_struct(variable, mtime, infile)
        if out_file.is_file() and not self.overwrite:
            return 0
        out_file.parent.mkdir(exist_ok=True, parents=True)
        with TemporaryDirectory(prefix='cdo') as td:
            temp_file = Path(td) / 'temp_file.nc'
            cmd = shlex.split(f'cdo -O -f nc4 --cmor chname,var{var_num},{variable} {infile} {temp_file}')
            res = run(cmd, check=True, stderr=PIPE, stdout=PIPE)
            self.correct_file(temp_file, out_file, variable, lookup=info)
        return 1

